from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from pydantic import BaseModel, ValidationError

from classification_model.config.core import config


def drop_na_inputs(*, input_data: pd.DataFrame) -> pd.DataFrame:
    """check model inputs for na values and filter"""
    validated_data = input_data.copy()
    new_vars_with_na = [
        var
        for var in config.model_config.features
        if validated_data[var].isnull().sum() > 0
    ]
    validated_data.dropna(subset=new_vars_with_na, inplace=True)

    return validated_data


def validate_inputs(*, input_data: pd.DataFrame) -> Tuple[pd.DataFrame, Optional[dict]]:
    """check model inputs for unprocessable values."""
    relevant_data = input_data[config.model_config.features].copy()
    validated_data = drop_na_inputs(input_data=relevant_data)
    errors = None

    try:
        # replace numpy nans so that pydantic can validate input
        MultipleCustomerDataInputs(
            inputs=validated_data.replace({np.nan: None}).to_dict(orient="records")
        )
    except ValidationError as error:
        errors = error.json()

    return validated_data, errors


class CustomerDataInputSchema(BaseModel):
    customerid: Optional[str]
    gender: Optional[str]
    seniorcitizen: Optional[str]
    partner: Optional[str]
    dependents: Optional[str]
    tenure: Optional[int]
    phoneservice: Optional[str]
    multiplelines: Optional[str]
    internetservice: Optional[str]
    onlinesecurity: Optional[str]
    onlinebackup: Optional[str]
    deviceprotection: Optional[str]
    techsupport: Optional[str]
    streamingtv: Optional[str]
    streamingmovies: Optional[str]
    contract: Optional[str]
    paperlessbilling: Optional[str]
    paymentmethod: Optional[str]
    monthlycharges: Optional[float]
    totalcharges: Optional[float]


class MultipleCustomerDataInputs(BaseModel):
    inputs: List[CustomerDataInputSchema]



import typing as t
from pathlib import Path

import joblib
import pandas as pd
from sklearn.pipeline import Pipeline

from classification_model import __version__ as _version
from classification_model.config.core import DATASET_DIR, TRAINED_MODEL_DIR, config


# this block of code returns a dataframe
def load_dataset(*, file_name: str) -> pd.DataFrame:

    df = pd.read_csv(Path(f"{DATASET_DIR}/{file_name}"))

    df.columns = df.columns.str.lower().str.replace(" ", "-")

    df = df.drop("customerid", axis=1)

    # format the values of our data by removing trailing spaces for
    # cat variable and equally changing them to lower case
    cat_var_formatting = [var for var in df.columns if df[var].dtype == "object"]

    for var in cat_var_formatting:
        if df[var].dtype == "object":
            df[var] = df[var].str.lower().str.replace(" ", "_")
        else:
            df[var] = df[var]

    # checking the data types of our feature columns and
    # possibily reformating some like total charges, senior citizen
    df.seniorcitizen = df.seniorcitizen.astype("object")
    df.totalcharges = pd.to_numeric(df.totalcharges, errors="coerce")

    df["churn"] = (df.churn == "yes").astype(int)

    df.dropna(inplace=True)

    return df


def save_pipeline(*, pipeline_to_persist: Pipeline) -> None:

    # prepare the versioned save file name
    save_file_name = f"{config.app_config.pipeline_save_file}{_version}.pkl"
    save_path = TRAINED_MODEL_DIR / save_file_name
    remove_old_pipeline(files_to_keep=[save_file_name])
    joblib.dump(pipeline_to_persist, save_path)


def load_pipeline(*, file_name: str) -> Pipeline:
    """load a persisted pipeline"""
    file_path = TRAINED_MODEL_DIR / file_name
    trained_model = joblib.load(filename=file_path)
    return trained_model


def remove_old_pipeline(*, files_to_keep: t.List[str]) -> None:
    do_not_delete = files_to_keep + ["__init__.py"]
    for model_file in TRAINED_MODEL_DIR.iterdir():
        if model_file.name not in do_not_delete:
            model_file.unlink()



import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler


# this class converts our dataset to a vector
class CategoricalEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, variables):

        self.variables = variables

    def fit(self, X, y=None):

        return self

    def transform(self, X):
        # #we create a copy of our dataframe
        X = X.copy()

        # create a dummy variables for our categories and combine them with the orignal dataframe
        X_new = pd.concat([X, pd.get_dummies(X[self.variables])], axis=1)

        # drop the original categorical variables
        X_new.drop(self.variables, axis=1, inplace=True)

        return X_new


class FeatureNormalizer(BaseEstimator, TransformerMixin):
    def __init__(self, variables, scalar=StandardScaler()):

        self.variables = variables
        self.scalar = scalar

    def fit(self, X, y=None):
        self.scalar.fit(X)
        return self

    def transform(self, X):
        # #we create a copy of our dataframe
        X = X.copy()

        scaled_X = pd.DataFrame(self.scalar.transform(X), columns=self.variables)

        return scaled_X


# select and return the features key to our modelling
class SelectedFeatures(BaseEstimator, TransformerMixin):
    def __init__(self, variables):

        self.variables = variables

    def fit(self, X, y=None):

        return self

    def transform(self, X):
        # #we create a copy of our dataframe
        X = X.copy()

        # select the key features
        X_new = X[self.variables]

        return X_new



import typing as t

# import numpy as np
import pandas as pd

from classification_model import __version__ as _version
from classification_model.config.core import config
from classification_model.processing.data_manager import load_pipeline
from classification_model.processing.validation import validate_inputs

pipeline_file_name = f"{config.app_config.pipeline_save_file}{_version}.pkl"
_customer_churn_pipeline = load_pipeline(file_name=pipeline_file_name)


def make_prediction(*, input_data: t.Union[pd.DataFrame, dict]) -> dict:
    """make a prediction using a saved model pipeline"""

    data = pd.DataFrame(input_data)
    validated_data, errors = validate_inputs(input_data=data)
    results = {"predictions": None, "version": _version, "errors": errors}

    if not errors:
        predictions = _customer_churn_pipeline.predict(
            X=validated_data[config.model_config.features]
        )
        results = {
            "predictions": [pred for pred in predictions],  # type: ignore
            "version": _version,
            "errors": errors,
        }
    return results



from config.core import config
from pipeline import customer_churn_pipeline
from processing.data_manager import load_dataset, save_pipeline
from sklearn.model_selection import train_test_split


def run_training() -> None:
    """train the model."""

    # read the training data
    data = load_dataset(file_name=config.app_config.training_data_file)

    x_train, x_test, y_train, y_test = train_test_split(
        data.drop([config.model_config.target], axis=1),  # predictors
        data[config.model_config.target],
        test_size=config.model_config.test_size,
        random_state=config.model_config.random_state,
    )

    # fit model
    customer_churn_pipeline.fit(x_train, y_train)

    # persist the trained model
    save_pipeline(pipeline_to_persist=customer_churn_pipeline)


if __name__ == "__main__":
    run_training()



import numpy as np

from classification_model.predict import make_prediction


def test_make_prediction(sample_input_data):
    # Given
    expected_first_prediction_value = 0
    expected_no_of_prediction = 2110

    # When
    result = make_prediction(input_data=sample_input_data)

    # Then
    predictions = result.get("predictions")
    assert isinstance(predictions, list)
    assert isinstance(predictions[0], np.int64)
    assert result.get("errors") is None
    assert len(predictions) == expected_no_of_prediction
    assert predictions[0] == expected_first_prediction_value



# import packages and custom feature
# for building our models
from sklearn.pipeline import Pipeline

# from sklearn.preprocessing import StandardScaler
from xgboost import XGBClassifier

from classification_model.config.core import config
from classification_model.processing import custom_feature as cf

# set up the pipeline
customer_churn_pipeline = Pipeline(
    [
        # == CATEGORICAL ENCODING ======
        # encode categorical variables using one hot encoding into k-1 variables
        (
            "categorical_encoder",
            cf.CategoricalEncoder(variables=config.model_config.cat_vars),
        ),
        # ==== FEATURE SELECTION ========
        (
            "selected_features",
            cf.SelectedFeatures(variables=config.model_config.selected_features),
        ),
        # ==== SCALING OUR data ========
        (
            "scaler",
            cf.FeatureNormalizer(variables=config.model_config.selected_features),
        ),
        (
            "xgb",
            XGBClassifier(
                eta=config.model_config.eta,
                alpha=config.model_config.alpha,
                random_state=config.model_config.random_state,
            ),
        ),
    ]
)
